classdef CircularArc
    %CIRCULARARC A circular arc in 2D.
    %   CircularArc(p_c, R, alpha0, alpha_del)
    % Properties:
    %   p_c         - center point of associated circle
    %   R           - radius of associated circle
    %   alpha0      - start angle of arc
    %   alpha_del   - (signed) central angle of arc
    %   L           - length
    % Methods:
    %   p(s)        - point at path parameter value s
    %   v(s)        - tangent vector at path parameter value s
    %   chi(s)      - course angle at path parameter value s
    %   kappa(s)    - (signed) curvature at path parameter s
    % The circular arc is arc length parameterized, so path parameter s is
    % limited to the interval [0, L].
    
    properties (SetAccess = immutable)
        p_c;        % Center point of associated circle
        R;          % Radius of associated circle
        alpha0;     % Start angle of arc
        alpha_del;  % (Signed) Central angle of arc
    end
    
    properties (Dependent)
        L;          % Length
    end
    
    methods
        function obj = CircularArc(p_c, R, alpha0, alpha_del)
            %CIRCULARARC CircularArc(p_c, R, alpha0, alpha_del)
            %   p_c         - center point of associated circle
            %   R           - radius of associated circle
            %   alpha0      - start angle of arc
            %   alpha_del   - (signed) central angle of arc
            
            p_c = p_c(:); % force column vector
            
            obj.checkPoint(p_c);
            obj.checkRadius(R);
            obj.checkStartAngle(alpha0);
            obj.checkCentralAngle(alpha_del);
            
            obj.p_c = p_c;
            obj.R = R;
            obj.alpha0 = alpha0;
            obj.alpha_del = alpha_del;
        end
        
        function L = get.L(obj)
            L = obj.R*abs(obj.alpha_del);
        end
        
        function p = p(obj, s)
            %P Point at path parameter value s
            %   p = P(s)
            
            obj.checkPathParameter(s);
            
            p = obj.p_c + obj.R*[cos(obj.alpha0 + s.*sign(obj.alpha_del)/obj.R);
                                 sin(obj.alpha0 + s.*sign(obj.alpha_del)/obj.R)];
        end
        
        function v = v(obj, s)
            %V Tangent vector at path parameter value s
            %   v = V(s)
            
            obj.checkPathParameter(s);
            
            v = sign(obj.alpha_del)*[-sin(obj.alpha0 + s.*sign(obj.alpha_del)/obj.R);
                                      cos(obj.alpha0 + s.*sign(obj.alpha_del)/obj.R)];
        end
        
        function chi = chi(obj, s)
            %CHI Course angle at path parameter value s
            %   chi = CHI(s)
            
            obj.checkPathParameter(s);
            
            chi = ssa(obj.alpha0 + s*sign(obj.alpha_del)/obj.R + sign(obj.alpha_del)*pi/2);
        end
        
        function kappa = kappa(obj, s)
            %KAPPA (Signed) Curvature at path parameter value s
            %   kappa = KAPPA(s)
            
            obj.checkPathParameter(s);
            
            kappa = ones(size(s))*sign(obj.alpha_del)/obj.R;
        end
    end
    
    methods (Access = private)
        function checkPoint(~, p)
            if size(p,1)~=2
                error('CircularArc:pointDimensionsWrong', 'Point must be 2x1 vector');
            end
        end
        
        function checkRadius(~, R)
            if R<=0
                error('CircularArc:radiusNotStrictlyPositive', 'Radius must be strictly positive');
            end
        end
        
        function checkStartAngle(~, alpha)
            if alpha>pi || alpha<-pi
                error('CircularArc:startAngleInvalid', 'Start angle must be in the interval [-pi, pi]');
            end
        end
        
        function checkCentralAngle(~, alpha_del)
            if alpha_del>2*pi || alpha_del<-2*pi
                error('CircularArc:centralAngleInvalid', 'Central angle must be in the interval [-2pi, 2pi]');
            end
        end
        
        function checkPathParameter(obj, s)
            if (any(s<0)) || (any(s>obj.L))
                error('CircularArc:pathParameterInvalid', 'Parametrization variable value must be in the interval [0, L]');
            end
        end
    end
end
