classdef EulerSpiral
    %EULERSPIRAL An EulerSpiral in 2D.
    %   EulerSpiral(p0, chi0, kappa0, kappa_del, w)
    % Properties:
    %   kappa_del   - (signed) curvature change
    %   w           - scale factor
    % Methods:
    %   p(s)        - point at path parameter value s
    %   v(s)        - tangent vector at path parameter value s
    %   chi(s)      - course angle at parameter value s
    %   kappa(s)    - (signed) curvature at path parameter s
    % The Euler spiral is arc length parameterized, so path parameter s is
    % limited to the interval [0, L].
    
    properties (GetAccess = private)
        chi0;       % Initial course angle
        kappa0;     % (Signed) Initial curvature
    end
    
    properties (SetAccess = immutable)
        p0;         % Start point
        p1;         % End point
        kappa_del;  % (Signed) Curvature change
        w;          % Scale factor
    end
    
    properties (Dependent)
        L;          % Length
    end
    
    methods
        function obj = EulerSpiral(p0, chi0, kappa0, kappa_del, w, p1)
            %EULERSPIRAL EulerSpiral(p0, chi0, kappa0, kappa_del, w, p1)
            %   p0          - start point
            %   chi0        - initial course angle
            %   kappa0      - (signed) initial curvature
            %   kappa_del   - (signed) curvature change
            %   w           - scale factor
            %   p1          - (optional) end point
            
            p0 = p0(:); % force column vector
            
            obj.checkPoint(p0);
            obj.checkCourseAngle(chi0);
            obj.checkScaleFactor(w);

            obj.p0 = p0;
            obj.chi0 = chi0;
            obj.kappa0 = kappa0;
            obj.kappa_del = kappa_del;
            obj.w = w;
            
            if nargin==6
                p1 = p1(:); % force column vector
                obj.p1 = p1;
            else
                obj.p1 = obj.p(obj.L);
            end
        end
        
        function L = get.L(obj)
            L = obj.w^2*abs(obj.kappa_del); 
        end
        
        function p = p(obj, s)
            %P Point at path parameter value s
            %   p = P(s)
            
            obj.checkPathParameter(s);
            
            p = zeros(2,length(s));
            for i=1:length(s)
                p(:,i) = obj.p0 + obj.w*[integral(@(tau) cos(obj.chi0 + tau*obj.kappa0*obj.w + 0.5*tau.^2*sign(obj.kappa_del)), 0, s(i)/obj.w);
                                         integral(@(tau) sin(obj.chi0 + tau*obj.kappa0*obj.w + 0.5*tau.^2*sign(obj.kappa_del)), 0, s(i)/obj.w)];
            end
        end
        
        function v = v(obj, s)
            %V Tangent vector at path parameter value s
            %   v = V(s)
            
            obj.checkPathParameter(s);
            
            v = [cos(obj.chi0 + s*obj.kappa0 + 0.5*s.^2*obj.kappa_del/obj.L);
                 sin(obj.chi0 + s*obj.kappa0 + 0.5*s.^2*obj.kappa_del/obj.L)];
        end
        
        function chi = chi(obj, s)
            %CHI Course angle at path parameter value s
            %   chi = CHI(s)
            
            obj.checkPathParameter(s);
            
            chi = ssa(obj.chi0 + s*obj.kappa0 + 0.5*s.^2*obj.kappa_del/obj.L);
        end
        
        function kappa = kappa(obj, s)
            %KAPPA Curvature at path parameter value s
            %   kappa = KAPPA(s)
            
            obj.checkPathParameter(s);
            
            kappa = obj.kappa0 + s*obj.kappa_del/obj.L;
        end
    end
    
    methods (Access = private)
        function checkPoint(~, p)
            if size(p,1)~=2
                error('EulerSpiral:pointDimensionsWrong', 'Point must be 2x1 vector');
            end
        end

        function checkCourseAngle(~, chi)
            if chi>pi || chi<-pi
                error('EulerSpiral:courseAngleInvalid', 'Initial course angle must be in the interval [-pi, pi]');
            end
        end

        function checkScaleFactor(~, w)
            if w<0
                error('EulerSpiral:scaleFactorNotStrictlyPositive', 'Scale factor must be strictly positive');
            end
        end

        function checkPathParameter(obj, s)
            if (any(s<0)) || (any(s>obj.L))
                error('EulerSpiral:pathParameterInvalid', 'Parametrization variable value must be in the interval [0, L]');
            end
        end
    end
end
