classdef LineSegment
    %LINESEGMENT A line segment in 2D.
    %   LineSegment(p0, p1)
    % Properties:
    %   L           - length
    % Methods:
    %   p(s)        - point at path parameter value s
    %   v(s)        - tangent vector at path parameter value s
    %   chi(s)      - course angle at path parameter value s
    %   kappa(s)    - (signed) curvature at path parameter s
    % The line segment is arc length parameterized, so path parameter s is limited
    % to the interval [0, L].
    
    properties (GetAccess = private)
        p0;     % Start point
        p1;     % End point
    end
    
    properties (Dependent)
        L;      % Length
    end
    
    methods
        function obj = LineSegment(p0, p1)
            %LINESEGMENT LineSegment(p0, p1)
            %   p0  - start point
            %   p1  - end point
            
            p0 = p0(:); % force column vector
            p1 = p1(:);
            
            obj.checkPoint(p0);
            obj.checkPoint(p1);
            
            obj.p0 = p0;
            obj.p1 = p1;
        end
        
        function L = get.L(obj)
            L = norm(obj.p1-obj.p0);
        end
        
        function p = p(obj, s)
            %P Point at path parameter value s
            %   p = P(s)
            
            obj.checkPathParameter(s);
            
            p = obj.p0 + s.*(obj.p1-obj.p0)/obj.L;
        end
        
        function v = v(obj, s)
            %V Tangent vector at path parameter value s
            %   v = V(s)
            
            obj.checkPathParameter(s);
            
            v = ones(size(s)).*(obj.p1-obj.p0)/obj.L;
        end
        
        function chi = chi(obj, s)
            %CHI Course angle at path parameter value s
            %   chi = CHI(s)
            
            obj.checkPathParameter(s);
            
            v = obj.v(s);
            chi = atan2(v(2,:), v(1,:));
        end
        
        function kappa = kappa(obj, s)
            %KAPPA (Signed) Curvature at path parameter value s
            %   kappa = KAPPA(s)
            
            obj.checkPathParameter(s);
            
            kappa = zeros(size(s));
        end
    end
    
    methods (Access = private)
        function checkPoint(~, p)
            if size(p,1)~=2
                error('LineSegment:pointDimensionsWrong', 'Point must be 2x1 vector');
            end
        end
        
        function checkPathParameter(obj, s)
            if (any(s<0)) || (any(s>obj.L))
                error('LineSegment:pathParameterInvalid', 'Parametrization variable value must be in the interval [0, L]');
            end
        end
    end
end
