classdef Spline2D
    %SPLINE2D Cubic spline approximation of a subpath in 2D.
    %   Spline2D(subpath)
    % Properties:
    %   ax          - spline coefficients for x coordinate
    %   ay          - spline coefficients for y coordinate
    %   L           - length
    % Methods:
    %   S(l)        - spline value at parameter value l
    %   S_der(l)    - first derivative spline value at parameter value l
    %   S_dder(l)   - second derivative spline value at parameter value l
    %   S_ddder(l)  - third derivative spline value at parameter value l
    % The spline is  (pseudo) arc length parameterized, so parameter l is limited
    % to the interval [0, L].
    
    properties
        ax;         % spline coefficients for x coordinate
        ay;         % spline coefficients for y coordinate
        L;          % spline length
    end
    
    methods
        function obj = Spline2D(subpath)
            %SPLINE2D Spline2D(subpath)
            %   subpath     - subpath of type LineSegment, CircularArc or EulerSpiral
            
            ax = zeros(1,4); % [ax0, ax1, ax2, ax3]
            ay = zeros(1,4);
            
            p0 = subpath.p(0);
            L = subpath.L;
            if isa(subpath, 'LineSegment')
                v = subpath.p(subpath.L) - p0;
                
                ax(1) = p0(1);
                ay(1) = p0(2);

                ax(2) = v(1)/L;
                ay(2) = v(2)/L;
            elseif isa(subpath, 'CircularArc')
                p1 = subpath.p(subpath.L);
                v_L = 2*norm(p1-p0)/(1+cos(abs(subpath.alpha_del)/2));
                v0 = subpath.v(0);
                v1 = subpath.v(subpath.L);
                v0 = v_L*v0/norm(v0);
                v1 = v_L*v1/norm(v1);
                
                
                ax(1) = p0(1);
                ay(1) = p0(2);

                ax(2) = v0(1)/L;
                ay(2) = v0(2)/L;

                ax(3) = (3*p1(1) - 3*p0(1) - 2*v0(1) - v1(1))/L^2;
                ay(3) = (3*p1(2) - 3*p0(2) - 2*v0(2) - v1(2))/L^2;

                ax(4) = (v0(1) + v1(1) + 2*p0(1) - 2*p1(1))/L^3;
                ay(4) = (v0(2) + v1(2) + 2*p0(2) - 2*p1(2))/L^3;
            elseif isa(subpath, 'EulerSpiral')
                % Rotation matrix
                Rz = @(a) [cos(a), -sin(a);
                           sin(a),  cos(a)];
                
                
                if subpath.kappa(0)==0
                    chi = subpath.chi(0);
                else
                    chi = subpath.chi(subpath.L);
                end

                p1 = subpath.p1;
                v_spiral_s_e = Rz(chi)'*(p1-p0);

                x_spiral = v_spiral_s_e(1);
                y_spiral = v_spiral_s_e(2);
                
                chi_delta = 0.5*subpath.L*subpath.kappa_del;
                
                ax(1) = p0(1);
                ay(1) = p0(2);

                if subpath.kappa(0)==0
                    ax(2) = (3*x_spiral - 4.5*y_spiral^2*subpath.kappa_del/(sin(chi_delta)^3) - 3*y_spiral/tan(chi_delta))/L;

                    ax(3) = (-3*x_spiral + 9*y_spiral^2*subpath.kappa_del/(sin(chi_delta)^3) + 3*y_spiral/tan(chi_delta))/(L^2);

                    ax(4) = (x_spiral - 4.5*y_spiral^2*subpath.kappa_del/(sin(chi_delta)^3))/(L^3);
                    ay(4) = y_spiral/(L^3);
                else
                    ax(2) = (3*y_spiral/tan(chi_delta))/L;
                    ay(2) = 3*y_spiral/L;

                    ax(3) = (4.5*y_spiral^2*subpath.kappa_del/(sin(chi_delta)^3) - 3*y_spiral/tan(chi_delta))/(L^2);
                    ay(3) = -3*y_spiral/(L^2);
                    
                    ax(4) = (x_spiral - 4.5*y_spiral^2*subpath.kappa_del/(sin(chi_delta)^3))/(L^3);
                    ay(4) = y_spiral/(L^3);
                end

                axy = Rz(chi)*[ax; ay];
                ax = [ax(1), axy(1,2:4)];
                ay = [ay(1), axy(2,2:4)];
            end
            
            obj.ax = ax;
            obj.ay = ay;
            obj.L = L;
        end
        
        function S = S(obj, l)
            %S Spline value at parameter value l
            %   S = S(l)
            
            obj.checkParameter(l);

            S = [obj.ax; obj.ay]*[ones(size(l)); l; l.^2; l.^3];
        end
        
        function S_der = S_der(obj, l)
            %S_der First derivative spline value at parameter value l
            %   S_der = S_der(l)
            
            obj.checkParameter(l);
            
            S_der = [obj.ax(2:4); obj.ay(2:4)]*[ones(size(l)); 2*l; 3*l.^2];
        end
        
        function S_dder = S_dder(obj, l)
            %S_dder Second derivative spline value at parameter value l
            %   S_dder = S_dder(l)
            
            obj.checkParameter(l);
            
            S_dder = [obj.ax(3:4); obj.ay(3:4)]*[2*ones(size(l)); 6*l];
        end
        
        function S_ddder = S_ddder(obj, l)
            %S_ddder Third derivative spline value at parameter value l
            %   S_ddder = S_ddder(l)
            
            obj.checkParameter(l);
            
            S_ddder = [obj.ax(4); obj.ay(4)]*6*ones(size(l));
        end
    end
    
    methods (Access = private)
        function checkParameter(obj, l)
            if (any(l<0)) || (any(l>obj.L))
                error('Spline2D:parameterInvalid', 'Parameterization variable value must be in the interval [0, L]');
            end
        end
    end
end
