classdef Dubins2D
    %DUBINS2D Interpolating 2D Dubins Path (G^1 continuous).
    %   Dubins2D(chi0, chi_n, R, p_wp)
    % Properties:
    %   table       - path table
    %   R           - radius of circular arcs
    %   p_wp        - waypoints
    %   L           - length
    %   debug       - debug data
    % Methods:
    %   p(s)        - point at path parameter value s
    %   v(s)        - tangent vector at path parameter value s
    %   chi(s)      - course angle at path parameter value s
    %   kappa(s)    - (signed) curvature at path parameter s
    % The path is arc length parameterized, so path parameter s is limited
    % to the interval [0, L].
    
    properties
        table;      % Path table
        R;          % Radius of circular arcs
        p_wp;       % Waypoints
        L;          % Length
        debug;      % Debug data
    end
    
    methods
        function obj = Dubins2D(chi0, chi_n, R, p_wp)
            %DUBINS2D Dubins2D(chi0, chi_n, R, p_wp)
            %   chi0    - initial course angle
            %   chi_n   - final course angle
            %   R       - radius of circular arcs
            %   p_wp    - waypoints
            
            n = length(p_wp);

            obj.checkCourseAngle(chi0);
            obj.checkCourseAngle(chi_n);
            obj.checkRadius(R);
            for i=1:n
                obj.checkWaypoint(p_wp(:,i));
            end
            
            % Data storage
            % It is not neccessary to store all of it, but is done for
            % debugging purposes
            v_in = zeros(2,n);
            v_out = zeros(2,n);
            v_wp = zeros(2,n);
            rho = zeros(1,n);
            
            v_wp_c = zeros(2,n);
            p_c = zeros(2,n);
            
            v_c_wo = zeros(2,n);
            v_c_po = zeros(2,n);
            p_wo = zeros(2,n);
            p_po = zeros(2,n);
            
            % Rotation matrix
            Rz = @(theta) [cos(theta), -sin(theta);
                           sin(theta),  cos(theta)];
            
            % Check assumption on waypoints distance
            wps_str = "";
            for i=1:n-1
                if norm(p_wp(:,i+1)-p_wp(:,i))<4*R
                    wps_str = wps_str + " " + i + "-" + (i+1);
                end
            end
            warning("Waypoints" + wps_str + " are closer than 4R. A solution may not exist.");

            % Initial course angle
            v_in(:,1) = Dubins2D.v_chi(chi0);
            v_out(:,1) = Dubins2D.v_p(p_wp(:,1), p_wp(:,2));
            v_wp(:,1) = v_in(:,1);
            rho(1) = Dubins2D.rho_v(v_in(:,1), v_out(:,1));
            p_wo(:,1) = p_wp(:,1);
            
            % Final course angle
            v_in(:,n) = Dubins2D.v_p(p_wp(:,n-1), p_wp(:,n));
            v_out(:,n) = Dubins2D.v_chi(chi_n);
            v_wp(:,n) = v_out(:,n);
            rho(n) = Dubins2D.rho_v(v_in(:,n), v_out(:,n));
            p_po(:,n) = p_wp(:,n);
            
            % Mean course change at waypoints (with piecewise linear path)
            for i=2:n-1
                v_in(:,i) = Dubins2D.v_p(p_wp(:,i-1), p_wp(:,i));
                v_out(:,i) = Dubins2D.v_p(p_wp(:,i), p_wp(:,i+1));
                v_wp(:,i) = Dubins2D.v_sum(v_in(:,i), v_out(:,i));
                rho(i) = Dubins2D.rho_v(v_in(:,i), v_out(:,i));
            end
            
            % Fix for zero course angle change waypoints
            for i=n-1:-1:1
                if rho(i)==0
                    v_wp(:,i-1) = v_out(:,i-1);
                    rho(i) = -rho(i+1);
                end
            end
            
            % Place circle centers
            for i=1:n
                v_wp_c(:,i) = Rz(rho(i)*pi/2)*v_wp(:,i);
                
                p_c(:,i) = p_wp(:,i) + R*v_wp_c(:,i);
            end
            
            % First wheel-over point and last pull-out point vectors
            v_c_wo(:,1) = Dubins2D.v_p(p_c(:,1), p_wo(:,1));
            v_c_po(:,n) = Dubins2D.v_p(p_c(:,n), p_po(:,n));
            
            % Wheel-over and pull-out points
            v_wo(:,1) = v_wp(:,1);
            v_po(:,n) = v_wp(:,n);
            for i=1:n-1
                v_c_c = p_c(:,i+1) - p_c(:,i);
                
                if rho(i)~=rho(i+1) % different consecutive rotations
                    if norm(0.5*v_c_c)<R
                        error("Dubins2D:noSolution", "Circle centers to waypoints "+i+" and "+(i+1)+" are closer than 2R when an external tangent is needed.\nNo solution exists.");
                    end
                    
                    beta = acos(R/(norm(0.5*v_c_c)));
                    
                    v_c_po(:,i) = Rz(-rho(i)*beta)*v_c_c/norm(v_c_c);
                    v_c_wo(:,i+1) = -v_c_po(:,i);
                    
                else % equivalent consecutive rotations                    
                    v_c_po(:,i) = Rz(-rho(i)*pi/2)*v_c_c/norm(v_c_c);
                    v_c_wo(:,i+1) = v_c_po(:,i);
                end
                
                p_po(:,i) = p_c(:,i) + R*v_c_po(:,i);
                p_wo(:,i+1) = p_c(:,i+1) + R*v_c_wo(:,i+1);
                
                v_po(:,i) = (p_wo(:,i+1)-p_po(:,i))/norm(p_wo(:,i+1)-p_po(:,i));
                v_wo(:,i+1) = v_po(:,i);
            end

            % Fix for full 360-degree turns
            for i=1:n
                
                rho_wo_wp = Dubins2D.rho_v(v_wo(:,i), v_wp(:,i));
                rho_wp_po = Dubins2D.rho_v(v_wp(:,i), v_po(:,i));
                
                while rho_wo_wp==-rho(i) || rho_wp_po==-rho(i)
                    if rho_wo_wp==-rho(i) && rho_wp_po==-rho(i)
                        rho(i) = -rho(i);
                    end
                    v_wp(:,i) = (v_wo(:,i)+v_po(:,i))/norm(v_wo(:,i)+v_po(:,i));
                    v_wp_c(:,i) = Rz(rho(i)*pi/2)*v_wp(:,i);
                    p_c(:,i) = p_wp(:,i) + R*v_wp_c(:,i);
                    
                    for j=i-1:i
                        v_c_c = p_c(:,j+1) - p_c(:,j);
                        if rho(j)~=rho(j+1) % different consecutive rotations
                            if norm(0.5*v_c_c)<R
                                error("Dubins2D:noSolution", "Circle centers to waypoints "+j+" and "+(j+1)+" are closer than 2R.\nNo solution exists.");
                            end

                            beta = acos(R/(norm(0.5*v_c_c)));

                            v_c_po(:,j) = Rz(-rho(j)*beta)*v_c_c/norm(v_c_c);
                            v_c_wo(:,j+1) = -v_c_po(:,j);

                        else % equivalent consecutive rotations                    
                            v_c_po(:,j) = Rz(-rho(j)*pi/2)*v_c_c/norm(v_c_c);
                            v_c_wo(:,j+1) = v_c_po(:,j);
                        end

                        p_po(:,j) = p_c(:,j) + R*v_c_po(:,j);
                        p_wo(:,j+1) = p_c(:,j+1) + R*v_c_wo(:,j+1);
                        
                        v_po(:,i) = (p_wo(:,i+1)-p_po(:,i))/norm(p_wo(:,i+1)-p_po(:,i));
                        v_wo(:,i+1) = v_po(:,i);
                    end
                    
                    rho_wo_wp = Dubins2D.rho_v(v_wo(:,i), v_wp(:,i));
                    rho_wp_po = Dubins2D.rho_v(v_wp(:,i), v_po(:,i));
                end
            end
            
            % Path table
            table = repmat(struct, [3*n-3, 1]);
            idx = 1;
            s = 0;
            for i=1:n
                if p_wo(:,i)~=p_wp(:,i) % circular arc from wheel-over to waypoint
                    alpha0 = Dubins2D.chi_v(v_c_wo(:,i));
                    alpha_del = mod(Dubins2D.chi_v(-v_wp_c(:,i)) - alpha0, rho(i)*2*pi);
                    circular = CircularArc(p_c(:,i), R, alpha0, alpha_del);

                    table(idx).subpath = circular;
                    table(idx).s0 = s;
                    s = s + circular.L;
                    table(idx).s1 = s;
                    table(idx).to_waypoint = i;
                    idx = idx + 1;
                end
                if p_wp(:,i)~=p_po(:,i) % circular arc from waypoint to pull-out
                    alpha0 = Dubins2D.chi_v(-v_wp_c(:,i));
                    alpha_del = mod(Dubins2D.chi_v(v_c_po(:,i)) - alpha0, rho(i)*2*pi);
                    circular = CircularArc(p_c(:,i), R, alpha0, alpha_del);

                    table(idx).subpath = circular;
                    table(idx).s0 = s;
                    s = s + circular.L;
                    table(idx).s1 = s;
                    table(idx).to_waypoint = 0;
                    idx = idx + 1;
                end                
                if i~=n % line segment from pull-out to wheel-over
                    line = LineSegment(p_po(:,i), p_wo(:,i+1));
                    table(idx).subpath = line;
                    table(idx).s0 = s;
                    s = s + line.L;
                    table(idx).s1 = s;
                    if p_wo(:,i+1)==p_wp(:,i+1)
                        table(idx).to_waypoint = i+1;
                    else
                        table(idx).to_waypoint = 0;
                    end
                    idx = idx + 1;
                end
            end
            
            obj.table = table(1:idx-1);
            obj.R = R;
            obj.p_wp = p_wp;
            obj.L = s;
            
            obj.debug.v_in = v_in;
            obj.debug.v_out = v_out;
            obj.debug.v_wp = v_wp;
            obj.debug.rho = rho;
            
            obj.debug.pc = p_c;
            
            obj.debug.p_wo = p_wo;
            obj.debug.p_po = p_po;
            obj.debug.v_wo = v_wo;
            obj.debug.v_po = v_po;
        end
        
        function p = p(obj, s)
            %P Point at path parameter value s
            %   p = P(s)
            
            obj.checkPathParameter(s);
            
            p = zeros(2,length(s));

            for i=1:length(obj.table)
                s_idx = find(s>=obj.table(i).s0 & s<obj.table(i).s1);
                if ~isempty(s_idx)
                    p(:,s_idx) = obj.table(i).subpath.p(s(s_idx)-obj.table(i).s0);
                    if s_idx(end)==length(s)
                        break;
                    end
                end
            end
            if s(end)==obj.L
               p(:,end) = obj.table(end).subpath.p(obj.table(end).s1-obj.table(end).s0);  
            end
        end
        
        function v = v(obj, s)
            %V Tangent vector at path parameter value s
            %   v = V(s)
            
            obj.checkPathParameter(s);
            
            v = zeros(2,length(s));
            for i=1:length(obj.table)
                s_idx = find(s>=obj.table(i).s0 & s<obj.table(i).s1);
                if ~isempty(s_idx)
                    v(:,s_idx) = obj.table(i).subpath.v(s(s_idx)-obj.table(i).s0);
                    if s_idx(end)==length(s)
                        break;
                    end
                end
            end
            if s(end)==obj.L
               v(:,end) = obj.table(end).subpath.v(obj.table(end).s1-obj.table(end).s0);  
            end
        end
        
        function chi = chi(obj, s)
            %CHI Course angle at path parameter value s
            %   chi = CHI(s)
            
            obj.checkPathParameter(s);
            
            chi = zeros(1,length(s));
            for i=1:length(obj.table)
                s_idx = find(s>=obj.table(i).s0 & s<obj.table(i).s1);
                if ~isempty(s_idx)
                    chi(:,s_idx) = obj.table(i).subpath.chi(s(s_idx)-obj.table(i).s0);
                    if s_idx(end)==length(s)
                        break;
                    end
                end
            end
            if s(end)==obj.L
               chi(:,end) = obj.table(end).subpath.chi(obj.table(end).s1-obj.table(end).s0);  
            end
        end
        
        function kappa = kappa(obj, s)
            %KAPPA (Signed) Curvature at path parameter value s
            %   kappa = KAPPA(s)
            
            obj.checkPathParameter(s);
            
            kappa = zeros(1,length(s));
            for i=1:length(obj.table)
                s_idx = find(s>=obj.table(i).s0 & s<obj.table(i).s1);
                if ~isempty(s_idx)
                    kappa(:,s_idx) = obj.table(i).subpath.kappa(s(s_idx)-obj.table(i).s0);
                    if s_idx(end)==length(s)
                        break;
                    end
                end
            end
            if s(end)==obj.L
               kappa(:,end) = obj.table(end).subpath.kappa(obj.table(end).s1-obj.table(end).s0);  
            end
        end
    end
    
    methods (Access = private)
        function checkWaypoint(~, p_wp)
            if size(p_wp,1)~=2
                error('Dubins2D:waypointWrongDimension', 'Waypoint must be 2x1 vector.');
            end
        end
        
        function checkCourseAngle(~, chi)
            if chi>pi || chi<-pi
                error('Dubins2D:courseAngleInvalid', 'Course angle must be in the interval [-pi, pi]');
            end
        end
        
        function checkRadius(~, R)
            if R<=0
                error('Dubins2D:radiusNotStrictlyPositive', 'Radius must be strictly positive');
            end
        end
        
        function checkPathParameter(obj, s)
            if (any(s<0)) || (any(s>obj.L))
                error('Dubins2D:pathParameterInvalid', 'Parametrization variable value must be in the interval [0, L].');
            end
        end
    end
    
    methods (Static, Access = private)
        function v = v_p(p0, p1)
            %V_P Unit vector from point p0 to p1
            %   v = V_P(p0, p1)
            
            v = (p1-p0)/norm(p1-p0);
        end
        
        function v = v_sum(v0, v1)
            %V_SUM Sum of vectors v0 and v1 as unit vector
            %   v = V_SUM(v0, v1)
            
            v = (v0+v1)/(norm(v0+v1));
        end
        
        function v = v_chi(chi)
            %V_CHI Unit vector with course angle chi
            %   v = V_CHI(chi)
            
            v = [cos(chi);
                 sin(chi)];
        end
        
        function chi = chi_v(v)
            %CHI_V Course angle of vector v
            %   chi = CHI_V(v)
            
            chi = atan2(v(2), v(1));
        end
        
        function rho = rho_v(v0, v1)
            %RHO_V Direction of rotation from vector v0 to v1
            %   rho = RHO_V(v0, v1)
            
            rho = -sign(v0(2)*v1(1) - v0(1)*v1(2));
        end
    end
end
