classdef ExtendedDubins2D
    %EXTENDEDDUBINS2D Interpolating 2D Extended Dubins Path (G^2 continuous)
    %   ExtendedDubins2D(chi0, chi_n, R, L_spiral, p_wp)
    % Properties:
    %   table        - path table
    %   R           - radius of circular arcs
    %   L_spiral    - length of Euler spirals
    %   p_wp        - waypoints
    %   L           - length
    %   debug       - debug data
    % Methods:
    %   p(s)        - point at path parameter value s
    %   v(s)        - tangent vector at path parameter value s
    %   chi(s)      - course angle at path parameter value s
    %   kappa(s)    - (signed) curvature at path parameter s
    % The path is arc length parameterized, so path parameter s is limited
    % to the interval [0, L].
    
    properties
        table;      % Path table
        R;          % Radius of circular arcs
        L_spiral;   % Length of Euler spirals
        p_wp;       % Waypoints
        L;          % Length
        debug;      % Debug data
    end
    
    methods
        function obj = ExtendedDubins2D(chi0, chi_n, R, L_spiral, p_wp)
            %EXTENDEDDUBINS2D ExtendedDubins2D(chi0, R, L_spiral, p_wp)
            %   chi0        - initial course angle
            %   chi_n       - final course angle
            %   R           - radius of circular arcs
            %   L_spiral    - length of Euler spirals
            %   p_wp        - waypoints
            
            n = length(p_wp);
            
            obj.checkCourseAngle(chi0);
            if nargin==5
                obj.checkCourseAngle(chi_n);
            end
            obj.checkRadius(R);
            obj.checkLengthEulerSpiral(L_spiral);
            for i=1:n
                obj.checkWaypoint(p_wp(:,i));
            end
            
            % Scale factor of Euler spirals
            w = sqrt(L_spiral*R);
            
            % Fresnel integrals
            % TODO: Offline computation
            v_spiral_s_e = fresnel_integrals(w, L_spiral);
            
            % Absolute value of change in course angle over the Euler spirals
            chi_del_spiral = 0.5*L_spiral/R;
            
            % Radius of outer circle
            R_spiral = R*cos(chi_del_spiral) + v_spiral_s_e(2);
            
            % Offset from wheel-over to wheel-over-start, and from
            % pull-out to pull-out-end
            L_offset = v_spiral_s_e(1) - R*sin(chi_del_spiral);

            % Data storage
            % It is not neccessary to store all of these, but is done for
            % debugging purposes
            v_in = zeros(2,n);
            v_out = zeros(2,n);
            v_wp = zeros(2,n);
            rho = zeros(1,n);
            
            v_wp_c = zeros(2,n);
            p_c = zeros(2,n);
            
            v_c_wo = zeros(2,n);
            v_c_po = zeros(2,n);
            p_wo = zeros(2,n);
            p_po = zeros(2,n);
            
            chi_wo = zeros(1,n);
            chi_po = zeros(1,n);
            p_wo_s = zeros(2,n);
            p_wo_e = zeros(2,n);
            p_po_s = zeros(2,n);
            p_po_e = zeros(2,n);
            
            % Rotation matrix and reflection matrix around x-axis
            Rz = @(chi) [cos(chi), -sin(chi);
                         sin(chi),  cos(chi)];
            Ref = @(rho) [1, 0;
                          0, rho];
            
            % Check assumption on waypoints distance
            wps_str = "";
            for i=1:n-1
                if norm(p_wp(:,i+1)-p_wp(:,i))<4*R
                    wps_str = wps_str + " " + i + "-" + (i+1);
                end
            end
            warning("Waypoints" + wps_str + " are closer than 4R. A solution may not exist.");

            % Initial
            v_in(:,1) = ExtendedDubins2D.v_chi(chi0);
            v_out(:,1) = ExtendedDubins2D.v_p(p_wp(:,1), p_wp(:,2));
            v_wp(:,1) = v_in(:,1);
            rho(1) = ExtendedDubins2D.rho_v(v_in(:,1), v_out(:,1));            
            p_wo_s(:,1) = p_wp(:,1);
            p_wo_e(:,1) = p_wo_s(:,1) + Rz(chi0)*Ref(rho(1))*v_spiral_s_e;
            p_c(:,1) = p_wo_e(:,1) + R*[cos(chi0+rho(1)*chi_del_spiral+0.5*pi);
                                        sin(chi0+rho(1)*chi_del_spiral+0.5*pi)];
            chi_wo(1) = chi0;
            p_wo(:,1) = p_wp(:,1);
            
            % Final
            v_in(:,n) = ExtendedDubins2D.v_p(p_wp(:,n-1), p_wp(:,n));
            if nargin==5
                v_out(:,n) = ExtendedDubins2D.v_chi(chi_n);
            else
                v_out(:,n) = v_in(:,n);
                chi_n = ExtendedDubins2D.chi_v(v_in(:,n));
            end
            v_wp(:,n) = v_out(:,n);
            rho(n) = ExtendedDubins2D.rho_v(v_in(:,n), v_out(:,n));
            p_po_e(:,n) = p_wp(:,n);
            p_po_s(:,n) = p_po_e(:,n) - Rz(chi_n)*Ref(-rho(n))*v_spiral_s_e;
            p_c(:,n) = p_po_s(:,n) + R*[cos(chi_n-rho(1)*chi_del_spiral+0.5*pi);
                                        sin(chi_n-rho(1)*chi_del_spiral+0.5*pi)];
            chi_po(n) = chi_n;
            p_po(:,n) = p_wp(:,n);

            % Mean course change at waypoints (with linear path)
            % and circle centers
            for i=2:n-1
                v_in(:,i) = ExtendedDubins2D.v_p(p_wp(:,i-1), p_wp(:,i));
                v_out(:,i) = ExtendedDubins2D.v_p(p_wp(:,i), p_wp(:,i+1));
                v_wp(:,i) = ExtendedDubins2D.v_sum(v_in(:,i), v_out(:,i));
                rho(i) = ExtendedDubins2D.rho_v(v_in(:,i), v_out(:,i));

                v_wp_c(:,i) = Rz(rho(i)*pi/2)*v_wp(:,i);
                p_c(:,i) = p_wp(:,i) + R*v_wp_c(:,i);
            end

            % Makeshift wheel-over and pull-out points
            for i=1:n-1
                v_c_c = p_c(:,i+1) - p_c(:,i);
                
                if rho(i)~=rho(i+1) % different consecutive rotations
                    if norm(0.5*v_c_c)<R
                        error("ExtendedDubins2D:noSolution", "Circle centers to waypoints "+i+" and "+(i+1)+" are closer than 2R when an external tangent is needed.\nNo solution exists.");
                    end
                    beta = acos(R_spiral/(norm(0.5*v_c_c)));
                    
                    v_c_po(:,i) = Rz(-rho(i)*beta)*v_c_c/norm(v_c_c);
                    v_c_wo(:,i+1) = -v_c_po(:,i);
                    
                else % equivalent consecutive rotations
                    v_c_po(:,i) = Rz(-rho(i)*pi/2)*v_c_c/norm(v_c_c);
                    v_c_wo(:,i+1) = v_c_po(:,i);
                end
                
                p_po(:,i) = p_c(:,i) + R_spiral*v_c_po(:,i);
                p_wo(:,i+1) = p_c(:,i+1) + R_spiral*v_c_wo(:,i+1);
                
                chi_po(i) = ExtendedDubins2D.chi_v(v_c_po(:,i)) + rho(i)*pi/2;
                chi_wo(i+1) = chi_po(i);
                
                if abs(ssa(chi_po(i)-chi_wo(i)))<2*chi_del_spiral
                   warning("The course change at waypoint " + i + " is less than the total course change of the entrance and exiting Euler spiral. A full 360-degree turn is performed at the waypoint."); 
                end
                
                p_po_e(:,i) = p_po(:,i) + L_offset*[cos(chi_po(i)); sin(chi_po(i))];
                p_po_s(:,i) = p_po_e(:,i) - Rz(chi_po(i))*Ref(-rho(i))*v_spiral_s_e;
                
                p_wo_s(:,i+1) = p_wo(:,i+1) - L_offset*[cos(chi_wo(i+1)); sin(chi_wo(i+1))];
                p_wo_e(:,i+1) = p_wo_s(:,i+1) + Rz(chi_wo(i+1))*Ref(rho(i+1))*v_spiral_s_e;
            end

            % Path table
            P = repmat(struct, [5*n-3, 1]);
            idx = 1;
            s = 0;
            
            % TODO: warning course change complete circle turn
            
            for i=1:n
                % Euler spiral from wheel-over-start to wheel-over-end
                spiral = EulerSpiral(p_wo_s(:,i), chi_wo(i), 0, rho(i)/R, w, p_wo_e(:,i));

                P(idx).subpath = spiral;
                P(idx).s0 = s;
                s = s + spiral.L;
                P(idx).s1 = s;
                P(idx).to_waypoint = 0;
                idx = idx + 1;
                
                if i==1 || i==n
                    % circular arc from wheel-over-end to pull-out-start
                    v_c_wo_e = ExtendedDubins2D.v_p(p_c(:,i), p_wo_e(:,i));
                    v_c_po_s = ExtendedDubins2D.v_p(p_c(:,i), p_po_s(:,i));
                    alpha0 = ExtendedDubins2D.chi_v(v_c_wo_e);
                    alpha_del = mod(ExtendedDubins2D.chi_v(v_c_po_s) - alpha0, rho(i)*2*pi);
                    circular = CircularArc(p_c(:,i), R, alpha0, alpha_del);

                    P(idx).subpath = circular;
                    P(idx).s0 = s;
                    s = s + circular.L;
                    P(idx).s1 = s;
                    P(idx).to_waypoint = 0;
                    idx = idx + 1;
                else
                    % circular arc from wheel-over-end to waypoint
                    v_c_wo_e = ExtendedDubins2D.v_p(p_c(:,i), p_wo_e(:,i));
                    alpha0 = ExtendedDubins2D.chi_v(v_c_wo_e);
                    alpha_del = mod(ExtendedDubins2D.chi_v(-v_wp_c(:,i)) - alpha0, rho(i)*2*pi);
                    circular = CircularArc(p_c(:,i), R, alpha0, alpha_del);

                    P(idx).subpath = circular;
                    P(idx).s0 = s;
                    s = s + circular.L;
                    P(idx).s1 = s;
                    P(idx).to_waypoint = i;
                    idx = idx + 1;

                    % circular arc from waypoint to pull-out-start
                    v_c_po_s = ExtendedDubins2D.v_p(p_c(:,i), p_po_s(:,i));
                    alpha0 = ExtendedDubins2D.chi_v(-v_wp_c(:,i));
                    alpha_del = mod(ExtendedDubins2D.chi_v(v_c_po_s) - alpha0, rho(i)*2*pi);
                    circular = CircularArc(p_c(:,i), R, alpha0, alpha_del);

                    P(idx).subpath = circular;
                    P(idx).s0 = s;
                    s = s + circular.L;
                    P(idx).s1 = s;
                    P(idx).to_waypoint = 0;
                    idx = idx + 1;
                end
                
                % Euler spiral from pull-out-start to pull-out-end
                spiral = EulerSpiral(p_po_s(:,i), chi_po(i)-rho(i)*chi_del_spiral, rho(i)/R, -rho(i)/R, w, p_po_e(:,i));

                P(idx).subpath = spiral;
                P(idx).s0 = s;
                s = s + spiral.L;
                P(idx).s1 = s;
                if i==n
                    P(idx).to_waypoint = i;
                else
                    P(idx).to_waypoint = 0;
                end
                idx = idx + 1;
                
                if i~=n
                    % line from pull-out-end to wheel-over-start
                    line = LineSegment(p_po_e(:,i), p_wo_s(:,i+1));
                    P(idx).subpath = line;
                    P(idx).s0 = s;
                    s = s + line.L;
                    P(idx).s1 = s;
                    P(idx).to_waypoint = 0;
                    idx = idx + 1;
                end
            end
            
            obj.table = P(1:idx-1);
            obj.R = R;
            obj.L_spiral = L_spiral;
            obj.p_wp = p_wp;
            obj.L = s;
            
            obj.debug.v_in = v_in;
            obj.debug.v_out = v_out;
            obj.debug.v_wp = v_wp;
            obj.debug.rho = rho;
            
            obj.debug.pc = p_c;
            
            obj.debug.p_po = p_po;
            obj.debug.p_wo = p_wo;
            obj.debug.chi_wo = chi_wo;
            obj.debug.chi_po = chi_po;
            
            obj.debug.p_wo_s = p_wo_s;
            obj.debug.p_wo_e = p_wo_e;
            obj.debug.p_po_s = p_po_s;
            obj.debug.p_po_e = p_po_e;
            
            obj.debug.R_spiral = R_spiral;
            obj.debug.L_offset = L_offset;
            obj.debug.chi_del_spiral = chi_del_spiral;
            obj.debug.v_spiral_s_e = v_spiral_s_e;
        end
        
        function p = p(obj, s)
            %P Point at path parameter value s
            %   p = P(s)
            
            obj.checkPathParameter(s);
            
            p = zeros(2,length(s));
            for i=1:length(obj.table)
                s_idx = find(s>=obj.table(i).s0 & s<obj.table(i).s1);
                if ~isempty(s_idx)
                    p(:,s_idx) = obj.table(i).subpath.p(s(s_idx)-obj.table(i).s0);
                    if s_idx(end)==length(s)
                        break;
                    end
                end
            end
            if s(end)==obj.L
               p(:,end) = obj.table(end).subpath.p(obj.table(end).s1-obj.table(end).s0);  
            end
        end
        
        function v = v(obj, s)
            %V Tangent vector at path parameter value s
            %   v = V(s)
            
            obj.checkPathParameter(s);
            
            v = zeros(2,length(s));
            for i=1:length(obj.table)
                s_idx = find(s>=obj.table(i).s0 & s<obj.table(i).s1);
                if ~isempty(s_idx)
                    v(:,s_idx) = obj.table(i).subpath.v(s(s_idx)-obj.table(i).s0);
                    if s_idx(end)==length(s)
                        break;
                    end
                end
            end
            if s(end)==obj.L
               v(:,end) = obj.table(end).subpath.v(obj.table(end).s1-obj.table(end).s0);  
            end
        end
        
        function chi = chi(obj, s)
            %CHI Course angle at path parameter value s
            %   chi = CHI(s)
            
            obj.checkPathParameter(s);
            
            chi = zeros(1,length(s));
            for i=1:length(obj.table)
                s_idx = find(s>=obj.table(i).s0 & s<obj.table(i).s1);
                if ~isempty(s_idx)
                    chi(:,s_idx) = obj.table(i).subpath.chi(s(s_idx)-obj.table(i).s0);
                    if s_idx(end)==length(s)
                        break;
                    end
                end
            end
            if s(end)==obj.L
               chi(:,end) = obj.table(end).subpath.chi(obj.table(end).s1-obj.table(end).s0);  
            end
        end
        
        function kappa = kappa(obj, s)
            %KAPPA (Signed) Curvature at path parameter value s
            %   kappa = KAPPA(s)
            
            obj.checkPathParameter(s);
            
            kappa = zeros(1,length(s));
            for i=1:length(obj.table)
                s_idx = find(s>=obj.table(i).s0 & s<obj.table(i).s1);
                if ~isempty(s_idx)
                    kappa(:,s_idx) = obj.table(i).subpath.kappa(s(s_idx)-obj.table(i).s0);
                    if s_idx(end)==length(s)
                        break;
                    end
                end
            end
            if s(end)==obj.L
               kappa(:,end) = obj.table(end).subpath.kappa(obj.table(end).s1-obj.table(end).s0);  
            end
        end
    end
    
    methods (Access = private)
        function checkWaypoint(~, p_wp)
            if size(p_wp,1)~=2
                error('ExtendedDubins2D:waypointWrongDimension', 'Waypoint must be 2x1 vector.');
            end
        end
        
        function checkCourseAngle(~, chi)
            if chi>pi || chi<-pi
                error('ExtendedDubins2D:courseAngleInvalid', 'Course angle must be in the interval [-pi, pi]');
            end
        end
        
        function checkRadius(~, R)
            if R<=0
                error('ExtendedDubins2D:radiusNotStrictlyPositive', 'Radius must be strictly positive');
            end
        end
        
        function checkLengthEulerSpiral(~, L)
            if L<0
                error('ExtendedDubins2D:lengthEulerSpiralNotStrictlyPositive', 'Length of Euler spiral must be strictly positive');
            end
        end
        
        function checkPathParameter(obj, s)
            if (any(s<0)) || (any(s>obj.L))
                error('ExtendedDubins2D:pathParameterInvalid', 'Parametrization variable value must be in the interval [0, L].');
            end
        end
    end
    
    methods (Static, Access = private)    
        function v = v_p(p0, p1)
            %V_P Unit vector from point p0 to p1
            %   v = V_P(p0, p1)
            
            v = (p1-p0)/norm(p1-p0);
        end
        
        function v = v_sum(v0, v1)
            %V_SUM Sum of vectors v0 and v1 as unit vector
            %   v = V_SUM(v0, v1)
            
            v = (v0+v1)/(norm(v0+v1));
        end
        
        function v = v_chi(chi)
            %V_CHI Unit vector with course angle chi
            %   v = V_CHI(chi)
            
            v = [cos(chi);
                 sin(chi)];
        end
        
        function chi = chi_v(v)
            %CHI_V Course angle of vector v
            %   chi = CHI_V(v)
            
            chi = atan2(v(2), v(1));
        end
        
        function rho = rho_v(v0, v1)
            %RHO_V Direction of rotation from vector v0 to v1
            %   rho = RHO_V(v0, v1)
            
            rho = -sign(v0(2)*v1(1) - v0(1)*v1(2));
        end
    end
end
