classdef SplinePath2D
    %SPLINEPATH2D Spline approximation of Dubins2D or ExtendedDubins2D
    %   SplinePath2D(path)
    % Properties:
    %   table           - spline path table
    %   p_wp            - waypoints
    %   L               - length [m]
    %   alpha_del_max   - max central angle in circular arcs [rad]
    % Methods:
    %   p(l)        - point at parameter value l
    %   chi(l)      - course angle at parameter value l
    %   kappa(l)    - curvature at parameter value l
    % The path is arc length parameterized, so path parameter l is limited
    % to the interval [0, L].
    
    properties
        table;                          % Spline path table
        p_wp;                           % Waypoints
        L;                              % Length
        alpha_del_max = deg2rad(45);    % Max central angle in circular arcs [rad]
    end

    methods
        function obj = SplinePath2D(path)
            %SPLINEPATH2D SplinePath2D(path)
            %   path    - path of type Dubins2D or ExtendedDubins2D
            
            % Split up circular arcs with central angle greater than
            % alpha_del_max to improve spline approximation
            table = repmat(struct, size(path.table));
            idx = 1;
            for i=1:length(table)
                if strcmp(class(path.table(i).subpath), "CircularArc") && abs(path.table(i).subpath.alpha_del)>obj.alpha_del_max
                    k = ceil(abs(path.table(i).subpath.alpha_del)/obj.alpha_del_max);
                    p_c = path.table(i).subpath.p_c;
                    R = path.table(i).subpath.R;
                    alpha0 = path.table(i).subpath.alpha0;
                    alpha_del = path.table(i).subpath.alpha_del/k;
                    s0 = path.table(i).s0;
                    for j=1:k
                        circular = CircularArc(p_c, R, alpha0, alpha_del);
                        table(idx).subpath = circular;
                        table(idx).s0 = s0;
                        table(idx).s1 = s0+circular.L;
                        table(idx).to_waypoint = 0;
                        idx = idx + 1;
                        alpha0 = ssa(alpha0+alpha_del);
                        s0 = s0 + circular.L;
                    end
                    table(idx-1).to_waypoint = path.table(i).to_waypoint;
                else
                    table(idx).subpath = path.table(i).subpath;
                    table(idx).s0 = path.table(i).s0;
                    table(idx).s1 = path.table(i).s1;
                    table(idx).to_waypoint = path.table(i).to_waypoint;
                    idx = idx + 1;
                end
            end
            
            spline_table = repmat(struct, size(table));
            for i=1:length(table)
                spline_table(i).spline = Spline2D(table(i).subpath);
                spline_table(i).l0 = table(i).s0;
                spline_table(i).l1 = table(i).s1;
                spline_table(i).to_waypoint = table(i).to_waypoint;
            end
            
            obj.table = spline_table;
            obj.L = spline_table(end).l1;
            obj.p_wp = path.p_wp;
        end
        
        function [S, S_der, S_dder, S_ddder] = spline_values(obj, l)
            
            
            obj.checkPathParameter(l);
            
            S = zeros(2,length(l));
            S_der = zeros(2,length(l));
            S_dder = zeros(2,length(l));
            S_ddder = zeros(2,length(l));

            for i=1:length(obj.table)
                l_idx = find(l>=obj.table(i).l0 & l<obj.table(i).l1);
                if ~isempty(l_idx)
                    S(:,l_idx) = obj.table(i).spline.S(l(l_idx)-obj.table(i).l0);
                    S_der(:,l_idx) = obj.table(i).spline.S_der(l(l_idx)-obj.table(i).l0);
                    S_dder(:,l_idx) = obj.table(i).spline.S_dder(l(l_idx)-obj.table(i).l0);
                    S_ddder(:,l_idx) = obj.table(i).spline.S_ddder(l(l_idx)-obj.table(i).l0);
                    if l_idx(end)==length(l)
                        break;
                    end
                end
            end
            if l(end)==obj.L
               S(:,end) = obj.table(end).spline.S(obj.table(end).l1-obj.table(end).l0);
               S_der(:,end) = obj.table(end).spline.S(obj.table(end).l1-obj.table(end).l0);
               S_dder(:,end) = obj.table(end).spline.S(obj.table(end).l1-obj.table(end).l0);
               S_ddder(:,end) = obj.table(end).spline.S(obj.table(end).l1-obj.table(end).l0);
            end
        end
        
        function p = p(obj, l)
            %P Point at parameter value l
            %   p = P(l)
                 
            obj.checkPathParameter(l);
            
            S = zeros(2,length(l));

            for i=1:length(obj.table)
                l_idx = find(l>=obj.table(i).l0 & l<obj.table(i).l1);
                if ~isempty(l_idx)
                    S(:,l_idx) = obj.table(i).spline.S(l(l_idx)-obj.table(i).l0);
                    if l_idx(end)==length(l)
                        break;
                    end
                end
            end
            if l(end)==obj.L
               S(:,end) = obj.table(end).spline.S(obj.table(end).l1-obj.table(end).l0);  
            end
            
            p = S;
        end
        
        function chi = chi(obj, l)
            %CHI Course angle at parameter value l
            %   chi = CHI(l)
            
            obj.checkPathParameter(l);
            
            S_der = zeros(2,length(l));

            for i=1:length(obj.table)
                l_idx = find(l>=obj.table(i).l0 & l<obj.table(i).l1);
                if ~isempty(l_idx)
                    S_der(:,l_idx) = obj.table(i).spline.S_der(l(l_idx)-obj.table(i).l0);
                    if l_idx(end)==length(l)
                        break;
                    end
                end
            end
            if l(end)==obj.L
               S_der(:,end) = obj.table(end).spline.S_der(obj.table(end).l1-obj.table(end).l0);  
            end
            
            chi = atan2(S_der(2,:), S_der(1,:));
        end
        
        function kappa = kappa(obj, l)
            %KAPPA Curvature at parameter value l
            %   kappa = KAPPA(l)
            
            obj.checkPathParameter(l);
            
            S_der = zeros(2,length(l));
            S_dder = zeros(2,length(l));

            for i=1:length(obj.table)
                l_idx = find(l>=obj.table(i).l0 & l<obj.table(i).l1);
                if ~isempty(l_idx)
                    S_der(:,l_idx) = obj.table(i).spline.S_der(l(l_idx)-obj.table(i).l0);
                    S_dder(:,l_idx) = obj.table(i).spline.S_dder(l(l_idx)-obj.table(i).l0);
                    if l_idx(end)==length(l)
                        break;
                    end
                end
            end
            if l(end)==obj.L
               S_der(:,end) = obj.table(end).spline.S_der(obj.table(end).l1-obj.table(end).l0);  
               S_dder(:,end) = obj.table(end).spline.S_dder(obj.table(end).l1-obj.table(end).l0);  
            end
            
            kappa = (S_der(1,:).*S_dder(2,:) - S_der(2,:).*S_dder(1,:))./((S_der(1,:).^2 + S_der(2,:).^2).^(3/2));
        end
        
        function kappa_der = kappa_der(obj, l)
            obj.checkPathParameter(l);
            
            S_der = zeros(2,length(l));
            S_dder = zeros(2,length(l));
            S_ddder = zeros(2,length(l));

            for i=1:length(obj.table)
                l_idx = find(l>=obj.table(i).l0 & l<obj.table(i).l1);
                if ~isempty(l_idx)
                    S_der(:,l_idx) = obj.table(i).spline.S_der(l(l_idx)-obj.table(i).l0);
                    S_dder(:,l_idx) = obj.table(i).spline.S_dder(l(l_idx)-obj.table(i).l0);
                    S_ddder(:,l_idx) = obj.table(i).spline.S_ddder(l(l_idx)-obj.table(i).l0);
                    if l_idx(end)==length(l)
                        break;
                    end
                end
            end
            if l(end)==obj.L
               S_der(:,end) = obj.table(end).spline.S_der(obj.table(end).l1-obj.table(end).l0);  
               S_dder(:,end) = obj.table(end).spline.S_dder(obj.table(end).l1-obj.table(end).l0);
               S_ddder(:,end) = obj.table(end).spline.S_ddder(obj.table(end).l1-obj.table(end).l0);
            end
            
            kappa_der = (S_der(1,:).*S_ddder(2,:)-S_der(2,:).*S_ddder(1,:))./((S_der(1,:).^2 + S_der(2,:).^2).^(3/2)) - ...
                3*(S_der(1,:).*S_dder(1,:)+S_der(2,:).*S_dder(2,:)).*(S_der(1,:).*S_dder(2,:)-S_der(2,:).*S_dder(1,:))./((S_der(1,:).^2 + S_der(2,:).^2).^(5/2));
        end
    end
    
    methods (Access = private)
        function checkPathParameter(obj, l)
            if (any(l<0)) || (any(l>obj.L))
                error('SplinePath2D:pathParameterInvalid', 'Parametrization variable value must be in the interval [0, L].');
            end
        end
    end
end
