% Example usage of the SplinePath2D class with the Dubins2D class
clear; close all; clc;

% UAV parameters
p0 = [-10; -1];         % Initial 2D position
chi0 = deg2rad(-45);    % Initial course angle [rad]
chi_n = deg2rad(90);    % Final course angle (at last waypoint) [rad]
Va = 18;                % Cruise airspeed [m/s]
phi_max = deg2rad(60);  % Max roll angle [rad]
p_des = deg2rad(120);   % Desired roll rate [rad/s]

g = 9.80665;                        % Gravitational acceleration [m/s^2]
chi_dot_max = g/Va*tan(phi_max);    % Turn rate with max roll angle in a coordinated turn [deg/s]
R = Va/chi_dot_max;                 % Radius of circular arcs [m]  

L_spiral = Va*(phi_max/p_des);          % Length of Euler spirals

% Constant altitude waypoints
load('WP.mat');
p_wp = [p0, WP];
n = length(p_wp);

% Spline approximation of Dubins2D
path = Dubins2D(chi0, chi_n, R, p_wp);
spline_path = SplinePath2D(path);

% What part of the path to plot below
l_start = spline_path.table(1).l0;
l_end = spline_path.table(end).l1;
l = l_start:0.01:l_end;
p_path = path.p(l);
chi_path = path.chi(l);
kappa_path = path.kappa(l);
p_spline = spline_path.p(l);
chi_spline = spline_path.chi(l);
kappa_spline = spline_path.kappa(l);

% XY-plane
figure(); hold on;
% Waypoints
for i=1:n
    plot(spline_path.p_wp(2,i), spline_path.p_wp(1,i), 'ks');
end
% Path
plot(p_path(2,:), p_path(1,:), 'k.', 'MarkerSize', 2);
plot(p_spline(2,:), p_spline(1,:), 'b.', 'MarkerSize', 2);
grid;
box on;
axis equal;
xlabel("y East [m]");
ylabel("x North [m]");
% Hacky legend
h = zeros(2, 1);
h(1) = plot(NaN, NaN, 'k');
h(2) = plot(NaN, NaN, 'b');
legend(h, "Path", "Spline");

figure();
% Position
subplot(4,1,1); hold on;
plot(l, p_path(1,:), 'k.', 'MarkerSize', 2);
plot(l, p_spline(1,:), 'b.', 'MarkerSize', 2);
for i=1:length(path.table)-1
    if ~strcmp(class(path.table(i).subpath), class(path.table(i+1).subpath))
        xline(path.table(i).s1, 'k:');
    end
end
ylabel("North [m]");
subplot(4,1,2); hold on;
plot(l, p_path(2,:), 'k.', 'MarkerSize', 2);
plot(l, p_spline(2,:), 'b.', 'MarkerSize', 2);
for i=1:length(path.table)-1
    if ~strcmp(class(path.table(i).subpath), class(path.table(i+1).subpath))
        xline(path.table(i).s1, 'k:');
    end
end
ylabel("East [m]");
% Course angle
subplot(4,1,3); hold on;
plot(l, chi_path, 'k.', 'MarkerSize', 2);
plot(l, chi_spline, 'b.', 'MarkerSize', 2);
for i=1:length(path.table)-1
    if ~strcmp(class(path.table(i).subpath), class(path.table(i+1).subpath))
        xline(path.table(i).s1, 'k:');
    end
end
%grid;
ylabel("Course angle [rad]");
% Curvature
subplot(4,1,4); hold on;
plot(l, kappa_path, 'k.', 'MarkerSize', 2);
plot(l, kappa_spline, 'b.', 'MarkerSize', 2);
for i=1:length(path.table)-1
    if ~strcmp(class(path.table(i).subpath), class(path.table(i+1).subpath))
        xline(path.table(i).s1, 'k:');
    end
end
%grid;
box on;
ylabel("Curvature [m^{-1}]");
xlabel("Path length [m]");
