classdef ExtendedDubins3D
    %EXTENDEDDUBINS3D Interpolating 3D Extended Dubins Path (C^2/G^2 continuous)
    %   ExtendedDubins3D(chi0, chi_n, R_xy, L_spiral, R_hs, gamma_max, p_wp)
    % Properties:
    %   path_xy     - ExtendedDubins2D path in horizontal X-Y plane
    %   path_hs     - Dubins2D path in vertical H-S_h plane
    %   R_xy        - radius of horizontal circular arcs
    %   L_spiral    - length of Euler spirals
    %   R_hs        - radius of vertical circular arcs
    %   gamma_max   - max flight path angle
    %   p_wp        - waypoints
    %   L           - length
    %   debug       - debug data
    % Methods:
    %   p(s)        - point at path parameter value s
    %   chi(s)      - course angle at path parameter value s
    %   gamma(s)    - flight path angle at path parameter value s
    %   kappa_h(s)  - horizontal curvature at path parameter s
    %   kappa_v(s)  - vertical curvature at path parameter s
    % The path is arc length parameterized, so path parameter s is limited
    % to the interval [0, L].
    
    properties
        path_xy;    % ExtendedDubins2D path in horizontal X-Y plane
        path_hs;    % Dubins2D path in vertical H-S_h plane
        R_xy;       % Radius of horizontal circular arcs
        L_spiral;   % Length of Euler spirals
        R_hs;       % Radius of vertical circular arcs
        gamma_max;  % Max flight path angle
        p_wp;       % Waypoints
        L;          % Length
        debug;      % Debug data
    end
    
    methods
        function obj = ExtendedDubins3D(chi0, chi_n, R_xy, L_spiral, R_hs, gamma_max, p_wp)
            %EXTENDEDDUBINS3D ExtendedDubins3D(chi0, chi_n, R_xy, L_spiral, R_hs, p_wp)
            %   chi0        - initial course angle
            %   chi_n       - final course angle
            %   R_xy        - radius of horizontal circular arcs
            %   L_spiral    - length of Euler spirals
            %   R_hs        - radius of vertical circular arcs
            %   gamma_max   - max flight path angle
            %   p_wp        - waypoints
            
            n = length(p_wp);
            
            p_wp_xy = p_wp(1:2, :);
            
            % ExtendedDubins2D in X-Y plane
            path_xy = ExtendedDubins2D(chi0, chi_n, R_xy, L_spiral, p_wp_xy);
            
            % Waypoints in H-S_h plane
            p_wp_hs = zeros(2, n);
            p_wp_hs(1,:) = -p_wp(3,:);
            p_wp_hs(2,1) = 0;
            for i=1:length(path_xy.table)
                if path_xy.table(i).to_waypoint
                    i_wp = path_xy.table(i).to_waypoint;
                    p_wp_hs(2,i_wp) = path_xy.table(i).s1;
                end
            end

            % Dubins2D in H-S_h plane
            gamma0 = 0;
            gamma_n = 0;
            chi0_v = 0.5*pi - gamma0;
            chi_n_v = 0.5*pi - gamma_n;
            path_hs = Dubins2D(chi0_v, chi_n_v, R_hs, p_wp_hs);

            % Fligth path angle to waypoints
            gamma = zeros(1,n);
            gamma(1) = 0;
            for i=1:length(path_hs.table)
                if path_hs.table(i).to_waypoint
                    i_wp = path_hs.table(i).to_waypoint;
                    gamma(i_wp) = 0.5*pi - path_hs.table(i).subpath.chi(0);
                end
            end
            
            % Add turns in horizontal plane if flight path angle constraint
            % is violated
            while any(gamma > gamma_max)
                for i=1:length(gamma)
                    if abs(gamma(i))>gamma_max
                        if i==2 % edge case
                            subpath = path_xy.table(2).subpath;
                            p_c = subpath.p_c;
                            R = subpath.R;
                            alpha0 = subpath.alpha0;
                            alpha_del = subpath.alpha_del;
                            circle = CircularArc(p_c, R, ssa(alpha0+alpha_del), sign(alpha_del)*2*pi);

                            tab.subpath = circle;
                            tab.s0 = path_xy.table(1).s1;
                            tab.s1 = tab.s0 + circle.L;
                            tab.to_waypoint = 0;
                            for k=2:length(path_xy.table)
                                path_xy.table(k).s0 = path_xy.table(k).s0 + circle.L;
                                path_xy.table(k).s1 = path_xy.table(k).s1 + circle.L;
                            end
                            path_xy.L = path_xy.L + circle.L;
                            path_xy.table = [path_xy.table(1); tab; path_xy.table(2:end)];
                        end
                        for j=1:length(path_xy.table)
                            if path_xy.table(j).to_waypoint==i-1
                                subpath = path_xy.table(j).subpath;
                                p_c = subpath.p_c;
                                R = subpath.R;
                                alpha0 = subpath.alpha0;
                                alpha_del = subpath.alpha_del;
                                circle = CircularArc(p_c, R, ssa(alpha0+alpha_del), sign(alpha_del)*2*pi);

                                tab.subpath = circle;
                                tab.s0 = path_xy.table(j).s1;
                                tab.s1 = tab.s0 + circle.L;
                                tab.to_waypoint = 0;
                                for k=j+1:length(path_xy.table)
                                    path_xy.table(k).s0 = path_xy.table(k).s0 + circle.L;
                                    path_xy.table(k).s1 = path_xy.table(k).s1 + circle.L;
                                end
                                path_xy.L = path_xy.L + circle.L;
                                path_xy.table = [path_xy.table(1:j); tab; path_xy.table(j+1:end)];
                                break;
                            end
                        end
                    end
                end
                
                % Waypoints in H-S_h plane
                p_wp_hs = zeros(2, n);
                p_wp_hs(1,:) = -p_wp(3,:);
                p_wp_hs(2,1) = 0;
                p_wp_hs(2,n) = path_xy.L;
                for i=1:length(path_xy.table)-1
                    if path_xy.table(i).to_waypoint
                        i_wp = path_xy.table(i).to_waypoint;
                        p_wp_hs(2,i_wp) = path_xy.table(i).s1;
                    end
                end
                
                % Dubins2D in H-S_h plane
                path_hs = Dubins2D(0.5*pi, 0.5*pi, R_hs, p_wp_hs);
                
                % Flight path angle to waypoints
                gamma = zeros(1,n);
                gamma(1) = 0;
                for i=1:length(path_hs.table)
                    if path_hs.table(i).to_waypoint
                        i_wp = path_hs.table(i).to_waypoint;
                        gamma(i_wp) = 0.5*pi - path_hs.table(i).subpath.chi(0);
                    end
                end
            end
            
            obj.path_xy = path_xy;
            obj.path_hs = path_hs;
            obj.p_wp = p_wp;
            obj.L = path_hs.L;
        end
        
        function p = p(obj, s)
            %P Point at path parameter value s
            %   p = P(s)
            
            obj.checkPathParameter(s);
            
            p = zeros(3,length(s));

            p_hs = obj.path_hs.p(s);
            h = p_hs(1,:);
            s_h = p_hs(2,:);
            
            p(3,:) = -h;
            p(1:2,:) = obj.path_xy.p(s_h);
        end
        
        function s_h = s_h(obj, s)
            %S_H Horizontal path length at path parameter value s
            
            obj.checkPathParameter(s);
            
            p_hs = obj.path_hs.p(s);
            
            s_h = p_hs(2,:);
        end
        
        function chi = chi(obj, s)
            %CHI Course angle at path parameter value s
            %   chi = CHI(s)
            
            obj.checkPathParameter(s);
            
            chi = obj.path_xy.chi(obj.s_h(s));
        end
        
        function gamma = gamma(obj, s)
            %GAMMA Flight path angle path parameter value s
            %   gamma = GAMMA(s)
            
            obj.checkPathParameter(s);
            
            gamma = 0.5*pi - obj.path_hs.chi(s);
        end
        
        function kappa_h = kappa_h(obj, s)
            %KAPPA_H Horizontal curvature at path parameter value s
            %   kappa_h = KAPPA_H(s)
            
            obj.checkPathParameter(s);
            
            kappa_h = obj.path_xy.kappa(obj.s_h(s));
        end
        
        function kappa_v = kappa_v(obj, s)
            %KAPPA_V Vertical curvature at path parameter value s
            %   kappa_v = KAPPA_V(s)
            
            obj.checkPathParameter(s);
            
            kappa_v = obj.path_hs.kappa(s);
        end
    end
    
    methods (Access = private)
        function checkPathParameter(obj, s)
            if (any(s<0)) || (any(s>obj.L))
                error('ExtendedDubins2D:pathParameterInvalid', 'Parametrization variable value must be in the interval [0, L].');
            end
        end
    end
end
