classdef SplinePath3D
    %SPLINEPATH3D Spline approximation of ExtendedDubins3D
    %   SplinePath3D(path)
    % Properties:
    %   spline_xy       - SplinePath2D in X-Y plane
    %   spline_hs       - SplinePath2D in H-S_h plane
    %   p_wp            - waypoints
    %   L               - length [m]
    %   alpha_del_max   - max central angle in circular arcs [rad]
    % Methods:
    %   p(l)            - point at parameter value l
    %   chi(l)          - course angle at parameter value l
    %   gamma(l)        - flight path angle at parameter value l
    %   kappa_h(l)      - horizontal curvature at parameter value l
    %   kappa_v(l)      - vertical curvature at parameter value l
    % The path is arc length parameterized, so path parameter s is limited
    % to the interval [0, L].
    
    properties
        spline_xy;                      % Spline path in X-Y plane table
        spline_hs;                      % Spline path in H-S_h plane table
        p_wp;                           % Waypoints
        L;                              % Length
        alpha_del_max = pi/8;           % Max central angle in circular arcs [rad]
    end
    
    methods
        function obj = SplinePath3D(path)
            %SPLINEPATH3D SplinePath3D(path)
            %   path    - path of type ExtendedDubins3D
    
            obj.spline_xy = SplinePath2D(path.path_xy);
            obj.spline_hs = SplinePath2D(path.path_hs);
            obj.L = path.path_hs.L;
            obj.p_wp = path.p_wp;
        end
        
        function [S, S_der, S_dder, S_ddder] = spline_values(obj, l)
            obj.checkPathParameter(l);

            [S_hs, S_hs_der, S_hs_dder, S_hs_ddder] = obj.spline_hs.spline_values(l);

            l_h = S_hs(2,:);

            [S_xy, S_xy_der, S_xy_dder, S_xy_ddder] = obj.spline_xy.spline_values(l_h);

            S(1:2,:) = S_xy;
            S(3,:) = -S_hs(1,:);

            S_der(1:2,:) = S_xy_der;
            S_der(3,:) = -S_hs_der(1,:);
            
            S_dder(1:2,:) = S_xy_dder;
            S_dder(3,:) = -S_hs_dder(1,:);
            
            S_ddder(1:2,:) = S_xy_ddder;
            S_ddder(3,:) = -S_hs_ddder(1,:);
        end
        
        function p = p(obj, l)
            %P Point at parameter value l
            %   p = P(l)
            
            obj.checkPathParameter(l);
            
            p = zeros(3,length(l));
            
            p_hs = obj.spline_hs.p(l);
            
            p(3,:) = -p_hs(1,:);
            l_h = p_hs(2,:);
            
            p_xy = obj.spline_xy.p(l_h);
            
            p(1:2,:) = p_xy;
        end
        
        function chi = chi(obj, l)
            %CHI Course angle at parameter value l
            %   chi = CHI(l)
            
            obj.checkPathParameter(l);
            
            p_hs = obj.spline_hs.p(l);
            l_h = p_hs(2,:);
            
            chi = obj.spline_xy.chi(l_h);
        end
        
        function gamma = gamma(obj, l)
            %GAMMA Flight path angle at parameter value l
            %   gamma = GAMMA(l)
            
            obj.checkPathParameter(l);
            
            gamma = 0.5*pi - obj.spline_hs.chi(l);
        end
        
        function kappa_h = kappa_h(obj, l)
            %KAPPA_H Horizontal curvature at parameter value l
            %   kappa_h = KAPPA_H(l)
            
            obj.checkPathParameter(l);
            
            p_hs = obj.spline_hs.p(l);
            l_h = p_hs(2,:);
            
            kappa_h = obj.spline_xy.kappa(l_h);
        end
        
        function kappa_v = kappa_v(obj, l)
            %KAPPA_V Vertical curvature at parameter value l
            %   kappa_v = KAPPA_V(l)
            
            obj.checkPathParameter(l);
            
            kappa_v = obj.spline_hs.kappa(l);
        end
        
        function kappa_h_der = kappa_h_der(obj, l)
            %KAPPA_H Horizontal curvature at parameter value l
            %   kappa_h = KAPPA_H(l)
            
            obj.checkPathParameter(l);
            
            p_hs = obj.spline_hs.p(l);
            l_h = p_hs(2,:);
            
            kappa_h_der = obj.spline_xy.kappa_der(l_h);
        end
    end
    
    methods (Access = private)
        function checkPathParameter(obj, s)
            if (any(s<0)) || (any(s>obj.L))
                error('SplinePath2D:pathParameterInvalid', 'Parametrization variable value must be in the interval [0, L].');
            end
        end
    end
end
