% Example script for the ExtendedDubins3D class
clear; close all; clc;

% NB: A bit time consuming as Fresnel integrals are solved for each s for
% plotting, the path must be spline approximated before being useful.

% UAV parameters
p0 = [-10; -1; -100];   % Initial 3D position
chi0 = deg2rad(-45);    % Initial course angle [rad]
chi_n = deg2rad(90);    % Final course angle (at last waypoint) [rad]
gamma0 = deg2rad(0);    % Initial flight path angle [rad]
gamma_n = deg2rad(0);   % Final fligth path angle [rad]
Va = 18;                % Cruise airspeed [m/s]
phi_max = deg2rad(60);  % Max roll angle [rad]
theta_max = deg2rad(30);    % Max pitch angle [rad]
p_des = deg2rad(120);   % Desired roll rate [rad/s]
q_des = deg2rad(60);   % Desired pitch rate [rad/s]

g = 9.80665;                        % Gravitational acceleration [m/s^2]
chi_dot_max = g/Va*tan(phi_max);    % Turn rate with max roll angle in a coordinated turn [deg/s]
R_xy = Va/chi_dot_max;              % Radius of horizontal circular arcs [m]  

L_spiral = Va*(phi_max/p_des);      % Length of Euler spirals

R_hs = Va/q_des;                    % Radius of vertical circular arcs [m]

% Constant altitude waypoints
load('WP3D_high2.mat');
p_wp = [p0, WP];
n = length(p_wp);

% ExtendedDubins3D
path = ExtendedDubins3D(chi0, chi_n, R_xy, L_spiral, R_hs, theta_max, p_wp);

% What part of the path to plot below
s_start = 0;
s_end = path.L;
s = s_start:0.01:s_end;
p = path.p(s);
s_h = path.s_h(s);

% X-Y plane
figure(); hold on;
% Waypoints
for i=1:n
    plot(path.path_xy.p_wp(2,i), path.path_xy.p_wp(1,i), 'ks');
end
% Subpaths
plot(p(2,:), p(1,:), 'b.', 'MarkerSize', 2);
% Initial and final course angle
quiver(path.path_xy.p_wp(2,1), path.path_xy.p_wp(1,1), Va*sin(chi0), Va*cos(chi0), 0, 'g--');
quiver(path.path_xy.p_wp(2,n), path.path_xy.p_wp(1,n), Va*sin(chi_n), Va*cos(chi_n), 0, 'r--');
grid;
box on;
axis equal;
xlabel("East y [m]");
ylabel("North x [m]");
% Hacky legend
h = zeros(3, 1);
h(1) = plot(NaN, NaN, 'b');
h(2) = plot(NaN, NaN, 'ks');
h(3) = plot(NaN, NaN, 'g--');
h(4) = plot(NaN, NaN, 'r--');
legend(h, "$\mathcal{P}$", "$\mathbf{p}_{wp}$", "$\chi_0$", "$\chi_n$", "Interpreter", "latex");

% H-S_h plane
figure(); hold on;
% Waypoints
for i=1:n
    plot(path.path_hs.p_wp(2,i), path.path_hs.p_wp(1,i), 'ks');
end
% Subpaths
plot(s_h, -p(3,:), 'b.', 'MarkerSize', 2);
% Initial and final flight path angle
quiver(path.path_hs.p_wp(2,1), path.path_hs.p_wp(1,1), Va*sin(0.5*pi-gamma0), Va*cos(0.5*pi-gamma0), 0, 'g--');
quiver(path.path_hs.p_wp(2,n), path.path_hs.p_wp(1,n), Va*sin(0.5*pi-gamma_n), Va*cos(0.5*pi-gamma_n), 0, 'r--');
grid;
box on;
axis equal;
xlabel("Horizontal path length s_h [m]");
ylabel("Altitude h [m]");
% Hacky legend
h = zeros(3, 1);
h(1) = plot(NaN, NaN, 'b');
h(2) = plot(NaN, NaN, 'ks');
h(3) = plot(NaN, NaN, 'g--');
h(4) = plot(NaN, NaN, 'r--');
legend(h, "$\mathcal{P}$", "$\mathbf{p}_{wp}$", "$\gamma_0$", "$\gamma_n$", "Interpreter", "latex");

% 3D
figure();
% Waypoints
plot3(p_wp(2,:), p_wp(1,:), -p_wp(3,:), 'ks');
hold on;
% Path
plot3(p(2,:), p(1,:), -p(3,:), 'b.', 'MarkerSize', 4);
grid;
box on;
axis equal;
xlabel("East y [m]");
ylabel("North x [m]");
zlabel("Altitude h [m]");
% Hacky legend
h = zeros(2, 1);
h(1) = plot(NaN, NaN, 'b');
h(2) = plot(NaN, NaN, 'ks');
legend(h, "$\mathcal{P}$", "$\mathbf{p}_{wp}$", "Interpreter", "latex");
zlim([50, 210]);

% Course angle
figure();
subplot(2,1,1); hold on;
s = 0:0.001:path.path_xy.L;
chi = path.chi(s);
plot(s, chi);
grid;
ylabel("Course angle \chi [rad]");
xlabel("Path length s [m]");
% Horizontal curvature
subplot(2,1,2);
kappa_h = path.kappa_h(s);
plot(s, kappa_h);
grid;
ylabel("Horizontal curvature \kappa_h [m^{-1}]");
xlabel("Path length s [m]");
ylim([-0.06, 0.06]);

% Flight path angle
figure();
subplot(2,1,1);
s = 0:0.001:path.path_hs.L;
gamma = 0.5*pi - path.path_hs.chi(s); % flight path angle = 0.5*pi - vertical course angle
plot(s, gamma);
for i=1:length(path.path_hs.table)-1
    if ~strcmp(class(path.path_hs.table(i).subpath), class(path.path_hs.table(i+1).subpath))
        xline(path.path_hs.table(i).s1, 'k:');
    end
end
grid;
ylabel("Flight path angle \gamma [rad]");
xlabel("Path length s [m]");
% Vertical curvature
subplot(2,1,2);
kappa_v = path.kappa_v(s);
plot(s, kappa_v, '.', 'MarkerSize', 3);
for i=1:length(path.path_hs.table)-1
    if ~strcmp(class(path.path_hs.table(i).subpath), class(path.path_hs.table(i+1).subpath))
        xline(path.path_hs.table(i).s1, 'k:');
    end
end
grid;
ylabel("Vertical curvature \kappa_v [m^{-1}]");
xlabel("Path length s [m]");
ylim([-0.065, 0.065]);

% Position
figure();
s = 0:0.01:path.L;
p = path.p(s);
subplot(3,1,1);
plot(s, p(1,:));
ylabel("x North [m]");
grid;
subplot(3,1,2);
plot(s, p(2,:));
grid;
ylabel("y East [m]");
subplot(3,1,3);
plot(s, -p(3,:));
grid;
ylabel("h Altitude [m]");
xlabel("s Path length [m]");
