% Example usage of the SplinePath3D class with the ExtendedDubins3D class
clear; close all; clc;

% UAV parameters
p0 = [-10; -1; -100];   % Initial 3D position
chi0 = deg2rad(-45);    % Initial course angle [rad]
chi_n = deg2rad(90);    % Final course angle (at last waypoint) [rad]
gamma0 = deg2rad(0);    % Initial flight path angle [rad]
gamma_n = deg2rad(0);   % Final flight path angle [rad]
Va = 18;                % Cruise airspeed [m/s]
phi_max = deg2rad(60);  % Max roll angle [rad]
theta_max = deg2rad(30); % Max pitch angle [rad]
p_des = deg2rad(120);   % Desired roll rate [rad/s]
q_des = deg2rad(60);   % Desired pitch rate [rad/s]

g = 9.80665;                        % Gravitational acceleration [m/s^2]
chi_dot_max = g/Va*tan(phi_max);    % Turn rate with max roll angle in a coordinated turn [deg/s]
R_xy = Va/chi_dot_max;              % Radius of horizontal circular arcs [m]  

L_spiral = Va*(phi_max/p_des);      % Length of Euler spirals

R_hs = Va/q_des;                    % Radius of vertical circular arcs [m]

% Constant altitude waypoints
load('WP3D_high2.mat');
p_wp = [p0, WP];
n = length(p_wp);

% ExtendedDubins3D
path = ExtendedDubins3D(chi0, chi_n, R_xy, L_spiral, R_hs, theta_max, p_wp);
spline_path = SplinePath3D(path);

% What part of the path to plot below
l_start = 0;
l_end = spline_path.L;
l = l_start:0.01:l_end;
p_path = path.p(l);
chi_path = path.chi(l);
gamma_path = path.gamma(l);
kappa_h_path = path.kappa_h(l);
kappa_v_path = path.kappa_v(l);
p_spline = spline_path.p(l);
chi_spline = spline_path.chi(l);
gamma_spline = spline_path.gamma(l);
kappa_h_spline = spline_path.kappa_h(l);
kappa_v_spline = spline_path.kappa_v(l);
kappa_h_der_spline = spline_path.kappa_h_der(l);


chi_dot_path = kappa_h_path*Va;
gamma_dot_path = kappa_v_path*Va;

chi_dot_spline = kappa_h_spline*Va;
gamma_dot_spline = -kappa_v_spline*Va;

% Roll angle
phi_path = atan(Va^2*kappa_h_path/g);
phi_spline = atan(Va^2*kappa_h_spline/g);


%p_rate_path = g*Va^2*kappa_h_der_path/(g^2 + Va^4*kappa_h_path^2);
q_path = gamma_dot_path + sin(phi_path).*Va.*kappa_h_path;
r_path = cos(phi_path).*Va.*kappa_h_path;


p_rate_spline = g*Va^3*kappa_h_der_spline./(g^2 + Va^4*kappa_h_spline.^2);
q_spline = gamma_dot_spline + sin(phi_spline).*Va.*kappa_h_spline;
r_spline = cos(phi_spline).*Va.*kappa_h_spline;

[S, S_der, S_dder, S_ddder] = spline_path.spline_values(l);

% 3D
figure();
% Waypoints
plot3(p_wp(2,:), p_wp(1,:), -p_wp(3,:), 'ks');
hold on;
% Path
plot3(p_spline(2,:), p_spline(1,:), -p_spline(3,:), 'b.', 'MarkerSize', 2);
grid;
box on;
axis equal;
xlabel("y East[m]");
ylabel("x North[m]");
zlabel("h Altitude [m]");
% Hacky legend
h = zeros(1, 1);
h(1) = plot(NaN, NaN, 'b');
legend(h, "Spline");


% Course angle
figure();
subplot(2,1,1); hold on;
%plot(l, chi_path, 'k.', 'MarkerSize', 2);
plot(l, chi_spline);
grid;
ylabel("Course angle \chi [rad]");
xlabel("s Path length [m]");
% Horizontal curvature
subplot(2,1,2); hold on;
%plot(l, kappa_h_path, 'k.', 'MarkerSize', 2);
plot(l, kappa_h_spline, '.', 'MarkerSize', 2);
grid;
ylabel("Horizontal curvature \kappa_h [m^{-1}]");
xlabel("Path length s [m]");
% % Hacky legend
% h = zeros(2, 1);
% h(1) = plot(NaN, NaN, 'k');
% h(2) = plot(NaN, NaN, 'b');
% legend(h, "Path", "Spline");
ylim([-0.06, 0.06]);

% Flight path angle
figure();
subplot(2,1,1); hold on;
%plot(l, gamma_path, 'k.', 'MarkerSize', 2);
plot(l, gamma_spline);
for i=1:length(path.path_hs.table)-1
    if ~strcmp(class(path.path_hs.table(i).subpath), class(path.path_hs.table(i+1).subpath))
        xline(path.path_hs.table(i).s1, 'k:');
    end
end
grid;
ylabel("Flight path angle \gamma [rad]");
xlabel("Path length s [m]");
% Vertical curvature
subplot(2,1,2); hold on;
%plot(l, kappa_v_path, 'k.', 'MarkerSize', 2);
plot(l, kappa_v_spline, '.', 'MarkerSize', 3);
for i=1:length(path.path_hs.table)-1
    if ~strcmp(class(path.path_hs.table(i).subpath), class(path.path_hs.table(i+1).subpath))
        xline(path.path_hs.table(i).s1, 'k:');
    end
end
grid;
ylim([-0.065, 0.065]);
ylabel("Vertical curvature \kappa_v [m^{-1}]");
xlabel("Path length s [m]");
% % Hacky legend
% h = zeros(2, 1);
% h(1) = plot(NaN, NaN, 'k');
% h(2) = plot(NaN, NaN, 'b');
% legend(h, "Path", "Spline");

% Position
figure();
subplot(3,1,1); hold on;
plot(l, p_path(1,:), 'k.', 'MarkerSize', 2);
plot(l, p_spline(1,:), 'b.', 'MarkerSize', 2);
ylabel("x North [m]");
grid;
subplot(3,1,2); hold on;
plot(l, p_path(2,:), 'k.', 'MarkerSize', 2);
plot(l, p_spline(2,:), 'b.', 'MarkerSize', 2);
grid;
ylabel("y East [m]");
subplot(3,1,3); hold on;
plot(l, -p_path(3,:), 'k.', 'MarkerSize', 2);
plot(l, -p_spline(3,:), 'b.', 'MarkerSize', 2);
grid;
ylabel("h Altitude [m]");
xlabel("s Path length [m]");
% Hacky legend
h = zeros(2, 1);
h(1) = plot(NaN, NaN, 'k');
h(2) = plot(NaN, NaN, 'b');
legend(h, "Path", "Spline");

% % Velocities
% figure();
% subplot(3,1,1); hold on;
% plot(l, S_der(1,:), 'b.', 'MarkerSize', 2);
% ylabel("S_x^\prime [m/s]");
% grid;
% subplot(3,1,2); hold on;
% plot(l, S_der(2,:), 'b.', 'MarkerSize', 2);
% grid;
% ylabel("S_y^\prime");
% subplot(3,1,3); hold on;
% plot(l, S_der(3,:), 'b.', 'MarkerSize', 2);
% grid;
% ylabel("S_z^\prime");
% xlabel("s Path length [m]");
% 
% % Acceleration
% figure();
% subplot(3,1,1); hold on;
% plot(l, S_dder(1,:), 'b.', 'MarkerSize', 2);
% ylabel("S_z^(\prime \prime) [m/s^2]");
% grid;
% subplot(3,1,2); hold on;
% plot(l, S_dder(2,:), 'b.', 'MarkerSize', 2);
% grid;
% ylabel("S_z^(\prime \prime)");
% subplot(3,1,3); hold on;
% plot(l, S_dder(3,:), 'b.', 'MarkerSize', 2);
% grid;
% ylabel("S_z^(\prime \prime)");
% xlabel("s Path length [m]");


% Roll angle
figure(); hold on;
%plot(l, phi_path, 'k.', 'MarkerSize', 2);
plot(l, phi_spline, 'b.', 'MarkerSize', 2);
grid;
ylabel("\phi Roll angle [rad]");
xlabel("s Path length [m]");
% % Hacky legend
% h = zeros(2, 1);
% h(1) = plot(NaN, NaN, 'k');
% h(2) = plot(NaN, NaN, 'b');
% legend(h, "Path", "Spline");

% Rates
figure();
subplot(3,1,1); hold on;
%plot(l, p_rate_path, 'k.', 'MarkerSize', 2);
plot(l, p_rate_spline, '.', 'MarkerSize', 3);
ylabel("Roll rate p [rad/s]");
grid;
subplot(3,1,2); hold on;
%plot(l, q_path, 'k.', 'MarkerSize', 2);
plot(l, q_spline, '.', 'MarkerSize', 3);
grid;
ylabel("Pitch rate q [rad/s]");
subplot(3,1,3); hold on;
%plot(l, r_path, 'k.', 'MarkerSize', 2);
plot(l, r_spline, '.', 'MarkerSize', 3);
grid;
ylabel("Yaw rate r [rad/s]");
xlabel("Path length s [m]");
% % Hacky legend
% h = zeros(2, 1);
% h(1) = plot(NaN, NaN, 'k');
% h(2) = plot(NaN, NaN, 'b');
% legend(h, "Path", "Spline");
